// plda/decodable-am-plda.cc

// Liang Lu (2014)

// See ../../COPYING for clarification regarding multiple authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.

#include <vector>
using std::vector;

#include "plda/decodable-am-plda-multi.h"

namespace kaldi {

BaseFloat DecodableAmPldaMulti::LogLikelihoodZeroBased(int32 frame, int32 pdf_id) {
  KALDI_ASSERT(frame >= 0 && frame < NumFrames());
  KALDI_ASSERT(pdf_id >= 0 && pdf_id < NumIndices());

  if (log_like_cache_[pdf_id].hit_time == frame) {
    return log_like_cache_[pdf_id].log_like;  // return cached value, if found
  }

  const VectorBase<BaseFloat> &data = feature_matrix_.Row(frame);
  // check if everything is in order
  if (acoustic_model_[0]->FeatureDim() != data.Dim()) {
    KALDI_ERR << "Dim mismatch: data dim = "  << data.Dim()
        << "vs. model dim = " << acoustic_model_[0]->FeatureDim();
  }

  int32 num_frames = acoustic_model_.size();	
  if (frame != previous_frame_) {  // Per-frame precomputation for PLDA.
    KALDI_ASSERT(frame < gselect_all_.size());
    
    //per_frame_vars_.resize(0);
    //PldaPerFrameDerivedVars *per_frame_vars = NULL;
    //per_frame_vars = new PldaPerFrameDerivedVars();
    
    gselect_ = gselect_all_[frame];
    acoustic_model_[0]->ComputePerFrameVars(data, gselect_, 0.0 /*FMLLR logdet*/, 
								      per_frame_vars_[0]);
   // per_frame_vars_.push_back(per_frame_vars);
    previous_frame_ = frame;
    
    for(int32 k = 1; k<num_frames; k+=2) {
      int32 i = frame - k;
      if(i < 0) i = 0;
    
   //   PldaPerFrameDerivedVars *per_frame_vars_bwd = NULL;
      gselect_ = gselect_all_[i];
    //  per_frame_vars_bwd = new PldaPerFrameDerivedVars();
      acoustic_model_[k]->ComputePerFrameVars(feature_matrix_.Row(i), gselect_, 0.0 /*FMLLR logdet*/, 
								      per_frame_vars_[k]);
    //  per_frame_vars_.push_back(per_frame_vars_bwd);
      
      i = frame + k;
      if(i>=feature_matrix_.NumRows()) i = feature_matrix_.NumRows() - 1;
      //PldaPerFrameDerivedVars *per_frame_vars_fwd = NULL;
      gselect_ = gselect_all_[i];
     // per_frame_vars_fwd = new PldaPerFrameDerivedVars();
      acoustic_model_[k+1]->ComputePerFrameVars(feature_matrix_.Row(i), gselect_, 0.0 /*FMLLR logdet*/, 
								      per_frame_vars_[k+1]);
     // per_frame_vars_.push_back(per_frame_vars_fwd);
    }
  }

  BaseFloat loglike = acoustic_model_[0]->LogLikelihood(*per_frame_vars_[0], pdf_id,
                                                    log_prune_);
						    
 // KALDI_LOG<<num_frames;
  BaseFloat alpha = 0.5;
  BaseFloat scale = 1.0; 
  for(int32 k = 1; k<num_frames; k+=2) {
  //  PldaPerFrameDerivedVars per_frame_vars;
 //   std::vector<int32> gselect;
 //   int32 i = frame - k;
//    if(i < 0) i = 0;
    
 //   gselect = gselect_all_[i];
 //   acoustic_model_[k]->ComputePerFrameVars(feature_matrix_.Row(i), gselect, 0.0 /*FMLLR logdet*/, 
//								      &per_frame_vars);
    loglike += alpha * acoustic_model_[k]->LogLikelihood(*per_frame_vars_[k], pdf_id,
                                                    log_prune_);
 //   KALDI_LOG<<loglike_tmp;

  //  i = frame + k;
  //  if(i>=feature_matrix_.NumRows()) i = feature_matrix_.NumRows() - 1;
    
   // gselect = gselect_all_[i];
  //  acoustic_model_[k+1]->ComputePerFrameVars(feature_matrix_.Row(i), gselect, 0.0,
//								      &per_frame_vars);
    loglike += alpha * acoustic_model_[k+1]->LogLikelihood(*per_frame_vars_[k+1], pdf_id,
                                                   log_prune_);
    scale += 2 * alpha;
    alpha *= 0.5;					    
  }
  
  loglike /= scale;
  if (KALDI_ISNAN(loglike) || KALDI_ISINF(loglike))
    KALDI_ERR << "Invalid answer (overflow or invalid variances/features?)";
  log_like_cache_[pdf_id].log_like = loglike;
  log_like_cache_[pdf_id].hit_time = frame;
  return loglike;

}

void DecodableAmPldaMulti::ResetLogLikeCache() {
  if (log_like_cache_.size() != acoustic_model_[0]->NumPdfs()) {
    log_like_cache_.resize(acoustic_model_[0]->NumPdfs());
  }
  vector<LikelihoodCacheRecord>::iterator it = log_like_cache_.begin(),
      end = log_like_cache_.end();
  for (; it != end; ++it) { it->hit_time = -1; }
}

}  // namespace kaldi
