// plda/mle-am-plda.cc

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.

//#include <algorithm>  // for std::max
//#include <string>
//#include <vector>

//#include "lfa/diag-fagmm.h"
#include "plda/mle-am-plda.h"

namespace kaldi {
using std::string;
using std::vector;

void MleAmPldaAccs::Write(std::ostream &out_stream, bool binary) const {
  uint32 tmp_uint32;
  
  WriteToken(out_stream, binary, "<AmPldaAcc>");
  WriteToken(out_stream, binary, "<NUMSTATES>");
  tmp_uint32 = static_cast<uint32>(num_states_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<NUMGaussians>");
  tmp_uint32 = static_cast<uint32>(num_gauss_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<FeatureDim>");
  tmp_uint32 = static_cast<uint32>(feature_dim_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<ChannelDim>");
  tmp_uint32 = static_cast<uint32>(chn_dim_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<PhoneDim>");
  tmp_uint32 = static_cast<uint32>(phn_dim_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<FLAGS>");
  WriteBasicType(out_stream, binary, flags_);
  
  if (!binary) out_stream << "\n";

  WriteToken(out_stream, binary, "<b_acc>");
  b_accumulator_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<var_acc>");
  var_accumulator_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<mean_x_acc>");
  mean_x_acc_.Write(out_stream, binary);
  
  if (U_accumulator_first_.size() != 0) {
 //   KALDI_ASSERT(U_accumulator_first_.size() != 0);
    WriteToken(out_stream, binary, "<U_first>");
    for (int32 i = 0; i < num_gauss_; i++) {
      U_accumulator_first_[i].Write(out_stream, binary);
    }
  }
  
  if (U_accumulator_second_.size() != 0) {
 //   KALDI_ASSERT(U_accumulator_second_.size() != 0);
    WriteToken(out_stream, binary, "<U_second>");
    for (int32 i = 0; i < num_gauss_; i++) {
      U_accumulator_second_[i].Write(out_stream, binary);
    }
  }
  
  if (G_accumulator_.size() != 0) {
 //   KALDI_ASSERT(G_accumulator_.size() != 0);
    WriteToken(out_stream, binary, "<G_first>");
    for (int32 i = 0; i < num_gauss_; i++) {
      G_accumulator_[i].Write(out_stream, binary);
    }
  }

  if (cov_x_acc_.size() != 0) {
    WriteToken(out_stream, binary, "<cov_x_acc>");
    for (int32 i = 0; i < num_gauss_; i++) {
      cov_x_acc_[i].Write(out_stream, binary);
    }
  }
  
  if (y_.size() != 0) {
//    KALDI_ASSERT(y_.size() != 0);
    WriteToken(out_stream, binary, "<y>");
    for (int32 j = 0; j < num_states_; j++) {
      y_[j].Write(out_stream, binary);
    }
  }
  
  if (gamma_.size() != 0) {
 //   KALDI_ASSERT(gamma_.size() != 0);
    WriteToken(out_stream, binary, "<gamma>");
    for (int32 j = 0; j < num_states_; j++) {
      gamma_[j].Write(out_stream, binary);
    }
  }
  
  WriteToken(out_stream, binary, "<total_like>");
  WriteBasicType(out_stream, binary, total_like_);

  WriteToken(out_stream, binary, "<total_frames>");
  WriteBasicType(out_stream, binary, total_frames_);
  
  WriteToken(out_stream, binary, "</AmPldaAcc>");
 }
 
 void MleAmPldaAccs::Read(std::istream &in_stream, bool binary, bool add){
   uint32 tmp_uint32;
   std::string token;
   //PldaUpdateFlagsType flags;
   //int32 num_gauss, phn_dim, chn_dim, feature_dim;
   
   ExpectToken(in_stream, binary, "<AmPldaAcc>");
   ExpectToken(in_stream, binary, "<NUMSTATES>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   num_states_ = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<NUMGaussians>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   num_gauss_ = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<FeatureDim>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   feature_dim_ = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<ChannelDim>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   chn_dim_ = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<PhoneDim>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   phn_dim_ = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<FLAGS>");
   ReadBasicType(in_stream, binary, &flags_);

   ReadToken(in_stream, binary, &token);
   while (token != "</AmPldaAcc>") {
      if(token == "<b_acc>") {
	//b_accumulator_.Resize(num_gauss_, feature_dim_);
        b_accumulator_.Read(in_stream, binary, add);
      } else if(token == "<var_acc>") {
	//var_accumulator_.Resize(num_gauss_, feature_dim_);
	var_accumulator_.Read(in_stream, binary, add);
      } else if(token == "<mean_x_acc>") {
	mean_x_acc_.Read(in_stream, binary, add);
      } else if (token == "<U_first>") {
	U_accumulator_first_.resize(num_gauss_);
        for (size_t i = 0; i < U_accumulator_first_.size(); i++) {
           U_accumulator_first_[i].Read(in_stream, binary, add);
        }
      } else if (token == "<U_second>") {
	U_accumulator_second_.resize(num_gauss_);
        for (size_t i = 0; i < U_accumulator_second_.size(); i++) {
          U_accumulator_second_[i].Read(in_stream, binary, add);
        }
      } else if (token == "<G_first>") {
	G_accumulator_.resize(num_gauss_);
	 for (size_t i = 0; i < G_accumulator_.size(); i++) {
           G_accumulator_[i].Read(in_stream, binary, add);
        }
      } else if (token == "<cov_x_acc>") {
	cov_x_acc_.resize(num_gauss_);
	for (size_t i = 0; i < cov_x_acc_.size(); i++) {
	  cov_x_acc_[i].Read(in_stream, binary, add);
	}
      } else if (token == "<y>") {
        y_.resize(num_states_);
      for (int32 j = 0; j < num_states_; j++) {
        y_[j].Read(in_stream, binary, add);
      }
    } else if (token == "<gamma>") {
      gamma_.resize(num_states_);
      for (int32 j = 0; j < num_states_; j++) {
        gamma_[j].Read(in_stream, binary, add);
      }
    } else if (token == "<total_like>") {
      double total_like;
      ReadBasicType(in_stream, binary, &total_like);
      if (add)
        total_like_ += total_like;
      else
        total_like_ = total_like;
    } else if (token == "<total_frames>") {
      double total_frames;
      ReadBasicType(in_stream, binary, &total_frames);
      if (add)
        total_frames_ += total_frames;
      else
        total_frames_ = total_frames;
    } else {
      KALDI_ERR << "Unexpected token '" << token << "' in model file ";
    }
      
      ReadToken(in_stream, binary, &token);
   }
 }

void MleAmPldaAccs::ResizeAccumulators(const AmPlda &model,
                                       PldaUpdateFlagsType flags) {
  num_states_ = model.NumPdfs();
  num_gauss_ = model.NumGauss();
  feature_dim_ = model.FeatureDim();
  phn_dim_ = model.PhoneDim();
  chn_dim_ = model.ChannelDim();

  if (flags & kPldaPhoneProjections) {
    G_accumulator_.resize(num_gauss_);
    for (int32 i = 0; i < num_gauss_; i++) {
      G_accumulator_[i].Resize(feature_dim_, phn_dim_);
    }
  } else {
    G_accumulator_.clear();
  }
  
  if (flags & kPldaChannelProjections) {
    U_accumulator_first_.resize(num_gauss_);
    U_accumulator_second_.resize(num_gauss_);
    for (int32 i = 0; i < num_gauss_; i++) {
      U_accumulator_first_[i].Resize(feature_dim_, chn_dim_);
      U_accumulator_second_[i].Resize(chn_dim_);
    }
  } else {
    U_accumulator_first_.clear();
    U_accumulator_second_.clear();
  }

  if(flags & kPldaChannelPrior) {
    mean_x_acc_.Resize(num_gauss_, chn_dim_);
    cov_x_acc_.resize(num_gauss_);
    for (int32 i = 0; i < num_gauss_; i++) {
      cov_x_acc_[i].Resize(chn_dim_);
    }
  }
  
  if (flags & kPldaVars) {
    var_accumulator_.Resize(num_gauss_, feature_dim_);
  } 

  if (flags & kPldaBias) {
    b_accumulator_.Resize(num_gauss_, feature_dim_);
  } 
  
  //occupancy count is always needed
  gamma_.resize(num_states_);
  total_frames_ = total_like_ = 0;
  for (int32 j = 0; j < num_states_; j++) {
    gamma_[j].Resize(model.NumSubstates(j), num_gauss_);
  }

  if (flags & kPldaPhoneVectors) {
    y_.resize(num_states_);
    for (int32 j = 0; j < num_states_; j++) {
      y_[j].Resize(model.NumSubstates(j), phn_dim_);
   //   y_[j].Resize(model.NumSubstates(j), feature_dim_);
    }
  } else {
    y_.clear();
  }
}

BaseFloat MleAmPldaAccs::Accumulate(const AmPlda &model,
                                    const PldaPerFrameDerivedVars &frame_vars,
                                    int32 state_index, BaseFloat weight,
                                    PldaUpdateFlagsType flags) {
  // Calculate Gaussian posteriors and collect statistics
  Matrix<BaseFloat> posteriors;
  BaseFloat log_like = model.ComponentPosteriors(frame_vars, state_index, &posteriors);
  posteriors.Scale(weight);
  BaseFloat count = AccumulateFromPosteriors(model, frame_vars, posteriors, state_index, flags);
  // Note: total_frames_ is incremented in AccumulateFromPosteriors().
  total_like_ += count * log_like;
  return log_like;
}


BaseFloat MleAmPldaAccs::AccumulateFromPosteriors(const AmPlda &model,
                                     const PldaPerFrameDerivedVars &frame_vars,
                                     const Matrix<BaseFloat> &posteriors,
                                     int32 j,
                                     PldaUpdateFlagsType flags) {  
  double tot_count = 0.0;
  bool usePostVar = true;
  const vector<int32> &gselect = frame_vars.gselect;
  // Intermediate variables
  Vector<BaseFloat> gammat(gselect.size());
  Vector<BaseFloat> xt_jmi(feature_dim_);
  Vector<BaseFloat> tmp_vec(feature_dim_);
  Vector<BaseFloat> U_xt(feature_dim_);
  Vector<BaseFloat> Gz_jm(feature_dim_);
  
  int32 num_substates = model.NumSubstates(j);
  for (int32 ki = 0; ki < static_cast<int32>(gselect.size()); ki++) {
    int32 i = gselect[ki];

    for (int32 m = 0; m < num_substates; m++) {
      // Eq. (39): gamma_{jmi}(t) = p (j, m, i|t)
      BaseFloat gammat_jmi = RandPrune(posteriors(ki, m), rand_prune_);
			
       // Accumulate statistics for non-zero gaussian posterior
      if (gammat_jmi != 0.0) {
        tot_count += gammat_jmi;
        if (flags & (kPldaPhoneVectors | kPldaChannelProjections |
                     kPldaVars| kPldaWeights | kPldaChannelPrior |kPldaBias)) {
          // Eq. (40): gamma_{jmi} = \sum_t gamma_{jmi}(t)
          gamma_[j](m, i) += gammat_jmi;
        }
        
        //compute the posterior mean of E[xt|yt, j, m, i]
        model.PosteriorChannelMean(frame_vars, j, m, i, &xt_jmi);
	if (flags & (kPldaPhoneVectors  | kPldaVars | kPldaBias)) {
          U_xt.AddMatVec(1.0, 
			model.ChannelProjection(i), 
			kNoTrans, 
			xt_jmi, 
			0.0);
	}
	
	if (flags & ( kPldaChannelProjections | kPldaVars | kPldaBias)) {
          Gz_jm.AddMatVec(1.0,
		        model.PhoneProjection(i), 
			kNoTrans, 
			model.StateVectors(j).Row(m), 
			0.0);
	}
	
        if (flags & kPldaChannelPrior) {
	  mean_x_acc_.Row(i).AddVec(gammat_jmi, xt_jmi);
	  cov_x_acc_[i].AddVec2(gammat_jmi, xt_jmi);
	}
	
        if (flags & kPldaPhoneVectors) {
	  y_[j].Row(m).AddVec(gammat_jmi, frame_vars.zti.Row(ki));
         }
  
         if (flags & kPldaChannelProjections) {
	  tmp_vec.SetZero();
	  tmp_vec.AddVec(gammat_jmi, frame_vars.xti.Row(ki));
	  tmp_vec.AddVec(-gammat_jmi, Gz_jm);
	  U_accumulator_first_[i].AddVecVec(1.0, tmp_vec, xt_jmi);
	  U_accumulator_second_[i].AddVec2(gammat_jmi, xt_jmi);
	  if(usePostVar) {
	    U_accumulator_second_[i].AddSp(gammat_jmi, SpMatrix<double> (model.InverseV(i)));
	  }
	 }
	 
	 if (flags & kPldaPhoneProjections) {
	  tmp_vec.SetZero();
	  tmp_vec.AddVec(gammat_jmi, frame_vars.xti.Row(ki));
	  tmp_vec.AddVec(-gammat_jmi, U_xt);
	  G_accumulator_[i].AddVecVec(1.0, tmp_vec, model.StateVectors(j).Row(m));
	}
	
	if (flags & kPldaBias) {
	 b_accumulator_.Row(i).AddVec(gammat_jmi, frame_vars.xt);
	 b_accumulator_.Row(i).AddVec(-gammat_jmi, Gz_jm);
	 b_accumulator_.Row(i).AddVec(-gammat_jmi, U_xt);
       }
       
       if(flags & kPldaVars) {
	tmp_vec.SetZero();
	tmp_vec.AddVec(1.0, frame_vars.xti.Row(ki));
	tmp_vec.AddVec(-1.0, U_xt);
	tmp_vec.AddVec(-1.0, Gz_jm);
        var_accumulator_.Row(i).AddVecVec(gammat_jmi, 
				    Vector<double> (tmp_vec), 
				    Vector<double> (tmp_vec), 
			            1.0);
        if(usePostVar) {
	  var_accumulator_.Row(i).AddVec(gammat_jmi, model.VarOffset().Row(i));
        }
       }
       
      }
    }
  }
  
   total_frames_ += tot_count;
   return tot_count;
  
}

// Compute the Q_{i} (Eq. 64)
void MleAmPldaUpdater::ComputeQ(const MleAmPldaAccs &accs,
                                const AmPlda &model,
                                std::vector< SpMatrix<double> > *Q) {
  Q->resize(accs.num_gauss_);
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    (*Q)[i].Resize(accs.phn_dim_);
    for (int32 j = 0; j < accs.num_states_; j++) {
      for (int32 m = 0; m < model.NumSubstates(j); m++) {
        if (accs.gamma_[j](m, i) > 0.0) {
          (*Q)[i].AddVec2(static_cast<BaseFloat>(accs.gamma_[j](m, i)),
                        model.v_[j].Row(m));
	  //Add the variance of sub-state vectors
      //    (*Q)[i].AddVec(static_cast<BaseFloat>(accs.gamma_[j](m, i)),
      //                  model.F_[j].Row(m));
        }
      }
    }
  }
}

void MleAmPldaUpdater::RenormalizeV(const MleAmPldaAccs &accs,
                                    AmPlda *model,
                                    const SpMatrix<double> &H_sm) {
  SpMatrix<double> Sigma(accs.phn_dim_);
  int32 count = 0;
  for (int32 j = 0; j < accs.num_states_; j++) {
    for (int32 m = 0; m < model->NumSubstates(j); m++) {
      count++;
      Sigma.AddVec2(static_cast<BaseFloat>(1.0), model->v_[j].Row(m));
    }
  }
  Sigma.Scale(1.0 / count);
  int32 fixed_eigs = Sigma.LimitCondDouble(update_options_.max_cond);
  if (fixed_eigs != 0) {
    KALDI_WARN << "Scatter of vectors v is poorly conditioned. Fixed up "
               << fixed_eigs << " eigenvalues.";
  }
  KALDI_LOG << "Eigenvalues of scatter of vectors v is : ";
  Sigma.PrintEigs("Sigma");
  if (!Sigma.IsPosDef()) {
    KALDI_LOG << "Not renormalizing v because scatter is not positive definite"
              << " -- maybe first iter?";
    return;
  }
  
   // Want to make variance of v unit and H_sm (like precision matrix) diagonal.
  TpMatrix<double> L(accs.phn_dim_);
  L.Cholesky(Sigma);
  TpMatrix<double> LInv(L);
  LInv.Invert();

  Matrix<double> tmpL(accs.phn_dim_, accs.phn_dim_);
  tmpL.CopyFromTp(L);

  SpMatrix<double> H_sm_proj(accs.phn_dim_);
  H_sm_proj.AddMat2Sp(1.0, tmpL, kTrans, H_sm, 0.0);
  // H_sm_proj := L^{T} * H_sm * L.
  // This is right because we would transform the vectors themselves
  // by L^{-1}, and H_sm is like the inverse of the vectors,
  // so it's {L^{-1}}^{-T} = L^T.

  Matrix<double> U(accs.phn_dim_, accs.phn_dim_);
  Vector<double> eigs(accs.phn_dim_);
  H_sm_proj.SymPosSemiDefEig(&eigs, &U, 1.0);  // 1.0 means no checking +ve def -> faster
  KALDI_LOG << "Note on the next diagnostic: the first number is generally not "
            << "that meaningful as it relates to the static offset";
  H_sm_proj.PrintEigs("H_sm_proj (Significance of dims in vector space.. note)");

  // Transform on vectors is U^T L^{-1}.
  // Why?  Because transform on H_sm is T =U^T L^T
  // and we want T^{-T} by normal rules of vector/covector and we
  // have (U^T L^T)^{-T} = (L U)^{-1} = U^T L^{-1}.
  Matrix<double> Trans(accs.phn_dim_, accs.phn_dim_);  // T^{-T}
  Matrix<double> tmpLInv(accs.phn_dim_, accs.phn_dim_);
  tmpLInv.CopyFromTp(LInv);
  Trans.AddMatMat(1.0, U, kTrans, tmpLInv, kNoTrans, 0.0);
  Matrix<double> TransInv(Trans);
  TransInv.Invert();  // T in above...

#ifdef KALDI_PARANOID
  {
    SpMatrix<double> H_sm_tmp(accs.phn_dim_);
    H_sm_tmp.AddMat2Sp(1.0, TransInv, kTrans, H_sm, 0.0);
    KALDI_ASSERT(H_sm_tmp.IsDiagonal(0.1));
  }
  {
    SpMatrix<double> Sigma_tmp(accs.phn_dim_);
    Sigma_tmp.AddMat2Sp(1.0, Trans, kNoTrans, Sigma, 0.0);
    KALDI_ASSERT(Sigma_tmp.IsUnit(0.1));
  }
#endif

  for (int32 j = 0; j < accs.num_states_; j++) {
    for (int32 m = 0; m < model->NumSubstates(j); m++) {
      Vector<double> tmp(accs.phn_dim_);
      tmp.AddMatVec(1.0, Trans, kNoTrans, Vector<double>(model->v_[j].Row(m)), 0.0);
      model->v_[j].Row(m).CopyFromVec(tmp);
    }
  }
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    Matrix<double> tmpG(accs.feature_dim_, accs.phn_dim_);
    // Multiplying on right not left so must not transpose TransInv.
    tmpG.AddMatMat(1.0, Matrix<double>(model->G_[i]), kNoTrans,
                   TransInv, kNoTrans, 0.0);
    model->G_[i].CopyFromMat(tmpG);
  }
  KALDI_LOG << "Renormalized subspace.";
}

// Compute H^{(sm)}, the "smoothing" matrices.
void MleAmPldaUpdater::ComputeSmoothingTerms(const MleAmPldaAccs &accs,
                                             const AmPlda &model,
                                             const vector<SpMatrix<double> > &H,
                                             SpMatrix<double> *H_sm) const {
  KALDI_ASSERT(H_sm != NULL);
  H_sm->Resize(accs.phn_dim_);
  Vector<double> gamma_i(accs.num_gauss_);

  double sum = 0.0;
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    if (gamma_i(i) > 0) {
      H_sm->AddSp(gamma_i(i), H[i]);
      sum += gamma_i(i);
    }
  }

  if (sum == 0.0) {
    KALDI_WARN << "Sum of counts is zero. Smoothing matrix set to unit";
    H_sm->SetUnit();  // arbitrary non-singular matrix
  } else {
    H_sm->Scale(1.0 / sum);
    Matrix<double> H_sm_old(*H_sm);
    int32 tmp = H_sm->LimitCondDouble(update_options_.max_cond_H_sm);
    if (tmp > 0) {
      KALDI_WARN << "Limited " << tmp << " eigenvalues of H_sm.";
    }
  }
}

double MleAmPldaUpdater::UpdateU(const MleAmPldaAccs &accs, AmPlda *model) {
  
  Matrix<double> U_i(accs.feature_dim_, accs.chn_dim_);
  SpMatrix<double> tmp_mat(accs.chn_dim_);
  SpMatrix<double> this_var(accs.feature_dim_);
  double tot_count = 0.0, tot_like_impr = 0.0;
  
  SolverOptions opts;
  opts.name = "U";
  opts.K = update_options_.max_cond;
  opts.eps = update_options_.epsilon;
    
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    double gamma_i = 0.0;
    for (int32 j = 0; j < accs.num_states_; j++)
      for (int32 m = 0; m < model->NumSubstates(j); m++)
        gamma_i += accs.gamma_[j](m, i);

    if (gamma_i < accs.feature_dim_) {
      KALDI_WARN << "For component " << i << ": not updating U due to very "
                 << "small count (=" << gamma_i << ").";
      continue;
    }
   
    tmp_mat.CopyFromSp(accs.U_accumulator_second_[i]);
    this_var.SetZero();
    this_var.AddVec(1.0, model->LambdaInv().Row(i));
    double impr =
        SolveQuadraticMatrixProblem(tmp_mat, accs.U_accumulator_first_[i],
                                    this_var, opts, &U_i);
  
  if (i < 10)
    KALDI_LOG<<"Trace UU^T before update "<<TraceMatMat(model->U_[i], model->U_[i], kTrans);
    model->U_[i].CopyFromMat(U_i);
  
  if (i < 10)  
    KALDI_LOG<<"Trace UU^T after update "<<TraceMatMat(model->U_[i], model->U_[i], kTrans);
  if (i < 10) {
      KALDI_VLOG(2) << "Objf impr for projection U for i = " << i << ", is "
                    << (impr/(gamma_i + 1.0e-20)) << " over " << gamma_i << " frames";
    }
    
    tot_count += gamma_i;
    tot_like_impr += impr;
  }
//  model->ComputeSigma();
  tot_like_impr /= (tot_count + 1.0e-20);
  KALDI_LOG << "Overall objective function improvement for model projections "
            << "U is " << tot_like_impr << " over " << tot_count << " frames";
  return tot_like_impr;
}

double MleAmPldaUpdater::UpdateG(const MleAmPldaAccs &accs, AmPlda *model) {
  Matrix<double> G_i(accs.feature_dim_, accs.phn_dim_);
  SpMatrix<double> this_var(accs.feature_dim_);
  double tot_count = 0.0, tot_like_impr = 0.0;
  
  SolverOptions opts;
  opts.name = "G";
  opts.K = update_options_.max_cond;
  opts.eps = update_options_.epsilon;
  
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    double gamma_i = 0.0;
    for (int32 j = 0; j < accs.num_states_; j++)
      for (int32 m = 0; m < model->NumSubstates(j); m++)
        gamma_i += accs.gamma_[j](m, i);

    if (gamma_i < accs.feature_dim_) {
      KALDI_WARN << "For component " << i << ": not updating G due to very "
                 << "small count (=" << gamma_i << ").";
      continue;
    }
   
    this_var.SetZero();
    this_var.AddVec(1.0, model->LambdaInv().Row(i));
 //  this_var.CopyFromSp(model->SigmaInv(i));
    double impr =
        SolveQuadraticMatrixProblem(Q_[i], accs.G_accumulator_[i],
                                    this_var, opts, &G_i);
    if (i < 10)			    
      KALDI_LOG<<"Trace GG^T before update "<<TraceMatMat(model->G_[i], model->G_[i], kTrans);
   // model->G_[i].Resize(feature_dim, state_dim);
    model->G_[i].CopyFromMat(G_i);
    
    if (i < 10)
    KALDI_LOG<<"Trace GG^T after update "<<TraceMatMat(model->G_[i], model->G_[i], kTrans);
    
    if (i < 10) {
      KALDI_VLOG(2) << "Objf impr for projection G for m = " << i << ", is "
                    << (impr/gamma_i + 1.0e-20) << " over " << gamma_i << " frames";
    }
    
    tot_count += gamma_i;
    tot_like_impr += impr;
  }
  
//   model->UpdateSigmaPLDA();
  return tot_like_impr;
}

double MleAmPldaUpdater::UpdateBias(const MleAmPldaAccs &accs, AmPlda *model) {
  double tot_count = 0.0, tot_like_impr = 0.0;
  Vector<BaseFloat> b_lambda_inv(accs.feature_dim_);
  
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    double gamma_i = 0.0;
    for (int32 j = 0; j < accs.num_states_; j++)
      for (int32 m = 0; m < model->NumSubstates(j); m++)
        gamma_i += accs.gamma_[j](m, i);

    if (gamma_i < accs.feature_dim_) {
      KALDI_WARN << "For component " << i << ": not updating b due to very "
                 << "small count (=" << gamma_i << ").";
      continue;
    }

    b_lambda_inv.AddVecVec(1.0, model->b_.Row(i), model->LambdaInv_.Row(i), 0.0);
    double auxf_old = -0.5 * gamma_i * VecVec(model->b_.Row(i), b_lambda_inv)
  			      + VecVec(b_lambda_inv, accs.b_accumulator_.Row(i));
  
  //  double auxf_old = -0.5 * gamma_i * VecSpVec(model->b_.Row(i), model->SigmaInv_[i], model->b_.Row(i))
//			      + VecSpVec(model->b_.Row(i), model->SigmaInv_[i], Vector<BaseFloat>(accs.b_accumulator_.Row(i)));
			      
    model->b_.Row(i).CopyFromVec(accs.b_accumulator_.Row(i));
    model->b_.Row(i).Scale(1.0/gamma_i);
    
    b_lambda_inv.AddVecVec(1.0, model->b_.Row(i), model->LambdaInv_.Row(i), 0.0);
    double auxf_new = -0.5 * gamma_i * VecVec(model->b_.Row(i), b_lambda_inv)
  			      + VecVec(b_lambda_inv, accs.b_accumulator_.Row(i));
			      
  //  double auxf_new = -0.5 * gamma_i * VecSpVec(model->b_.Row(i), model->SigmaInv_[i], model->b_.Row(i))
//			      + VecSpVec(model->b_.Row(i), model->SigmaInv_[i], Vector<BaseFloat>(accs.b_accumulator_.Row(i)));
			      
    if (i < 10) {
      KALDI_VLOG(2) << "Objf impr bias b for i = " << i << ", is "
                    << ((auxf_new - auxf_old)/gamma_i + 1.0e-20) << " over " << gamma_i << " frames";
    }
    tot_count += gamma_i;
    tot_like_impr += auxf_new - auxf_old;
  }
 
//   model->ComputeGconsts(); 
  return tot_like_impr;
}

double MleAmPldaUpdater::UpdateVars(const MleAmPldaAccs &accs, AmPlda *model) {
  int32 tot_floored = 0, gauss_floored = 0;
  Vector<double> var_floor(accs.feature_dim_);
  Vector<BaseFloat> tmp_vec(accs.feature_dim_);
  Vector<BaseFloat> gamma_vec(accs.num_gauss_);
  double tot_count = 0.0, tot_like_impr = 0.0;
  
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    double gamma_i = 0.0;
    for (int32 j = 0; j < accs.num_states_; j++)
      for (int32 m = 0; m < model->NumSubstates(j); m++)
        gamma_i += accs.gamma_[j](m, i);
  
    gamma_vec(i) = gamma_i;
    var_floor.AddVec(1.0, accs.var_accumulator_.Row(i));
  }
  
  //KALDI_LOG<<gamma_vec<<gamma_vec.Sum();
  
  var_floor.Scale(update_options_.cov_floor/gamma_vec.Sum());
  
 // KALDI_LOG<<var_floor;
  
  for(int32 i=0; i<accs.num_gauss_; i++) {
    tmp_vec.CopyRowFromMat(model->LambdaInv_, i);
    tmp_vec.ApplyLog();
    double auxf_old =0.5 * gamma_vec(i) * tmp_vec.Sum()
			    -0.5 * VecVec(model->LambdaInv_.Row(i), accs.var_accumulator_.Row(i));
			    
    model->LambdaInv_.Row(i).CopyFromVec(accs.var_accumulator_.Row(i));
    model->LambdaInv_.Row(i).Scale(1.0/gamma_vec(i));
    
    int32 floored;
    floored = model->LambdaInv_.Row(i).ApplyFloor(Vector<BaseFloat>(var_floor)); 
    if (floored != 0) {
       tot_floored += floored;
       gauss_floored++;
       KALDI_WARN<<"floored "<<floored <<" variance values for component " <<i; 
     }
     
    model->LambdaInv_.Row(i).InvertElements();
    tmp_vec.CopyRowFromMat(model->LambdaInv_, i);
    tmp_vec.ApplyLog();
    double auxf_new =0.5 * gamma_vec(i) * tmp_vec.Sum()
			    -0.5 * VecVec(model->LambdaInv_.Row(i), accs.var_accumulator_.Row(i));

     if (i < 10) {
      KALDI_VLOG(2) << "Objf impr for variance  i = " << i << ", is "
                    << ((auxf_new - auxf_old)/gamma_vec(i) + 1.0e-20) << " over " << gamma_vec(i) << " frames";
    }
    tot_count += gamma_vec(i);
    tot_like_impr += auxf_new - auxf_old;
  }

  if (tot_floored > 0)
    KALDI_WARN << tot_floored << " variances floored in " << gauss_floored
               << " Gaussians.";
	       
  return tot_like_impr;
   
}

double MleAmPldaUpdater::UpdateWeights(const MleAmPldaAccs &accs, AmPlda *model){
  Vector<double> gamma_j(accs.num_gauss_);
  double tot_gamma = 0.0, objf_impr = 0.0;
  
  for (int32 j = 0; j < accs.num_states_; j++) {
    int32 num_substates = model->NumSubstates(j);
    Vector<double> occs(num_substates);
    
    gamma_j.SetZero();
    for (int32 m = 0; m < num_substates; m++) {
      for (int32 i = 0; i < accs.num_gauss_; i++)
        gamma_j(i) += accs.gamma_[j](m, i);
      
      occs(m) = accs.gamma_[j].Row(m).Sum();  // \sum_i gamma_{jmi}
    }
    
    //update sub-state weights
    for (int32 m = 0; m < num_substates; m++) {
      double cur_weight = model->c_[j](m);
      if (cur_weight <= 0) {
        KALDI_WARN << "Zero or negative weight, flooring";
        cur_weight = 1.0e-10;  // future work(arnab): remove magic numbers
      }
      model->c_[j](m) = (occs(m) + update_options_.tau_c) 
			    / (occs.Sum() + num_substates * update_options_.tau_c);
      objf_impr += log(model->c_[j](m) / cur_weight) * occs(m);
    }
    
    double tot_gamma_j = gamma_j.Sum();
    tot_gamma += tot_gamma_j;
    int32 floored = 0;
    for (int32 i = 0; i < accs.num_gauss_; i++) {
      double cur_weight = model->w_(j,i);
       if (gamma_j(i) > update_options_.min_gaussian_occupancy)
      model->w_(j,i) = gamma_j(i)/tot_gamma_j;
      else {
	floored++;
	model->w_(j,i) = update_options_.min_gaussian_weight;
       }
       
       objf_impr += log(model->w_(j,i) / cur_weight) * gamma_j(i);
    }
    
    if (j<10)
      KALDI_LOG << "The number floored weights of state j = " << j << " is " << floored;	   
  }
  
  objf_impr /= (tot_gamma + 1.0e-20);
  KALDI_LOG << "**Overall objf impr for c is " << objf_impr << ", over "
            << tot_gamma << " frames.";
  return objf_impr;
  }

void MleAmPldaUpdater::UpdateChannelPrior(const MleAmPldaAccs &accs, AmPlda *model) {
  int32 tot_floored = 0, gauss_floored = 0;
  SpMatrix<double> var_floor(accs.chn_dim_);
  Vector<BaseFloat> gamma_vec(accs.num_gauss_);
  Matrix<BaseFloat> old_mean(accs.num_gauss_, accs.chn_dim_);
  
  old_mean.CopyFromMat(model->mean_x_);
  for (int32 i = 0; i < accs.num_gauss_; i++) {
    double gamma_i = 0.0;
    for (int32 j = 0; j < accs.num_states_; j++)
      for (int32 m = 0; m < model->NumSubstates(j); m++)
        gamma_i += accs.gamma_[j](m, i);
  
    gamma_vec(i) = gamma_i;
  }
  
  //compute the covariance flooring matrix
  for(int32 i=0; i<accs.num_gauss_; i++) {
    var_floor.AddSp(1.0, accs.cov_x_acc_[i]);
    var_floor.AddVec2(gamma_vec(i), old_mean.Row(i));
    var_floor.AddVecVec(-1.0, accs.mean_x_acc_.Row(i), Vector<double>(old_mean.Row(i)));
 //   var_floor.AddVecVec(-1.0, model->mean_x_.Row(i), accs.mean_x_acc_.Row(i));
  }
  
  var_floor.Scale(update_options_.cov_floor/gamma_vec.Sum());
  var_floor.LimitCondDouble(update_options_.max_cond);
  
  //update prior mean of x
  for(int32 i=0; i<accs.num_gauss_; i++) {
     if (gamma_vec(i) < accs.feature_dim_) {
      KALDI_WARN << "For component " << i << ": not updating prior mean of x due to very "
                 << "small count (=" << gamma_vec(i) << ").";
      continue;
     }
    model->mean_x_.Row(i).CopyFromVec(accs.mean_x_acc_.Row(i));
    model->mean_x_.Row(i).Scale(1.0/gamma_vec(i));
  }
  
  for(int32 i=0; i<accs.num_gauss_; i++) {
    if (gamma_vec(i) < accs.feature_dim_) {
      KALDI_WARN << "For component " << i << ": not updating prior covariance of x due to very "
                 << "small count (=" << gamma_vec(i) << ").";
      continue;
     }
    model->CovInv_x_[i].CopyFromSp(accs.cov_x_acc_[i]);
    model->CovInv_x_[i].AddVec2(gamma_vec(i), old_mean.Row(i));
    model->CovInv_x_[i].AddVecVec(-1.0, Vector<BaseFloat>(accs.mean_x_acc_.Row(i)), old_mean.Row(i));
    model->CovInv_x_[i].Scale(1.0/gamma_vec(i));
    
    int32 floored;
    floored = model->CovInv_x_[i].ApplyFloor(SpMatrix<BaseFloat>(var_floor)); 
    if (floored != 0) {
       tot_floored += floored;
       gauss_floored++;
       KALDI_WARN<<"floored "<<floored <<" variance values for component " <<i; 
     }
     
    model->CovInv_x_[i].InvertDouble();
  }

  if (tot_floored > 0)
    KALDI_WARN << tot_floored << " variances floored in " << gauss_floored
               << " Gaussians.";
	       
 // return tot_like_impr;
}

double MleAmPldaUpdater::UpdatePhoneVectors(const MleAmPldaAccs &accs,
                                                   AmPlda *model,
                                                   const std::vector< SpMatrix<double> > &H) {
  KALDI_LOG << "Updating phone vectors (and checking auxiliary function)";

  double tot_count = 0.0,  tot_auxf_impr = 0.0;  // sum over all states
      
  SolverOptions opts;
  opts.name = "v";
  opts.K = update_options_.max_cond;
  opts.eps = update_options_.epsilon;
  
  //compuate variance of state vectors
 // model->F_.resize(model->NumPdfs());
  for (int32 j = 0; j < accs.num_states_; j++) {
    
 //   model->F_[j].Resize(model->NumSubstates(j), model->PhoneDim());
    for (int32 m = 0; m < model->NumSubstates(j); m++) {
      double gamma_jm = accs.gamma_[j].Row(m).Sum();
      SpMatrix<double> H_jm(accs.phn_dim_);  // = \sum_i \gamma_{jmi} H_i

    //  H_jm.SetUnit();
      for (int32 i = 0; i < accs.num_gauss_; i++) {
        double gamma_jmi = accs.gamma_[j](m, i);
        if (gamma_jmi != 0.0)
          H_jm.AddSp(gamma_jmi, H[i]);
      }

//      KALDI_LOG<<H_jm;
//      model->F_[j].Row(m).CopyDiagFromSp(SpMatrix<BaseFloat>(H_jm));
//      model->F_[j].Row(m).InvertElements();
      
      Vector<double> v_jm(model->v_[j].Row(m));
      Vector<double> y_jm(accs.y_[j].Row(m));  
      double auxf_impr = SolveQuadraticProblem(H_jm, y_jm, opts, &v_jm);
      model->v_[j].Row(m).CopyFromVec(v_jm);
      
      tot_count += gamma_jm;
      tot_auxf_impr += auxf_impr;      

      if (j < 10 && m < 2) {
        KALDI_LOG << "Objf impr for j = " << (j) << " m = " << (m) << " is " << auxf_impr/gamma_jm
                          << " over " << gamma_jm << " frames";
      }
    }
  }

  tot_auxf_impr /= (tot_count + 1.0e-20);
  KALDI_LOG << "**Overall objf impr for v is " << tot_auxf_impr
            << " over " << tot_count << " frames";
  // Choosing to return actual likelihood impr here.
  return tot_auxf_impr;
}

double MleAmPldaUpdater::Update(const MleAmPldaAccs &accs, 
					      AmPlda *model, 
					      PldaUpdateFlagsType flags) {
                      
  KALDI_ASSERT((flags & (kPldaPhoneVectors | kPldaPhoneProjections |
                         kPldaChannelProjections | kPldaVars | kPldaChannelPrior|
                         kPldaBias | kPldaWeights)) != 0);
			 
 // if (flags & ~accs.flags_)
 //   KALDI_ERR << "Flags in argument do not match the active accumulators";
   
  if (flags & kPldaPhoneProjections)
    ComputeQ(accs, *model, &Q_);
  
  // quantities used in vector updates...
  vector< SpMatrix<double> > H;
  // "smoothing" matrices, weighted sums of above.
  SpMatrix<double> H_sm;
  if (flags & kPldaPhoneVectors) 
    model->ComputeH(&H);
  
  if (update_options_.renormalize_V)
    ComputeSmoothingTerms(accs, *model, H, &H_sm);
			  
  double tot_impr = 0.0;
   
  if (flags & kPldaPhoneVectors) {
    tot_impr += UpdatePhoneVectors(accs, model, H);
  }
  
  if (flags &  kPldaChannelProjections) {
     tot_impr += UpdateU(accs, model);
  }
  
  if (flags & kPldaBias) {
    tot_impr += UpdateBias(accs, model);
  }
  
  if (flags & kPldaWeights) {
    tot_impr += UpdateWeights(accs, model);
  }
 
  if (flags &  kPldaPhoneProjections) {
     tot_impr += UpdateG(accs, model);
  }
  
  if (flags & kPldaVars) {
    tot_impr += UpdateVars(accs, model);
  }

/*
  if(flags & kPldaChannelPrior) {
    UpdateChannelPrior(accs, model);  
  } else {
    model->InitChannelPrior();
  }
  */

  if(flags & (kPldaChannelProjections | kPldaVars |kPldaChannelPrior)) {
     model->ComputeV();  //update the matrix inv_V
  }
  
  if (update_options_.renormalize_V)
    RenormalizeV(accs, model, H_sm);
    
  KALDI_LOG << "*Overall auxf improvement, combining all parameters, is "
            << tot_impr;

  KALDI_LOG << "***Overall data likelihood is "
            << (accs.total_like_/accs.total_frames_)
            << " over " << (accs.total_frames_) << " frames.";
  
  model->ComputeSigmaInv();
  model->ComputeNormalizers();  // So that the model is ready to use.
  return tot_impr;
}




}  // End of namespace kaldi
