

#include "base/kaldi-common.h"
#include "util/common-utils.h"
#include "plda/fa-gmm.h"
#include "plda/mle-fa-gmm.h"

int main(int argc, char *argv[]) {
  try {
    using namespace kaldi;

    const char *usage =
        "Estimate a factor analysis (FA) GMM from the accumulated stats.\n"
        "Usage:  fa-gmm-est [options] <model-in> <stats-in> <model-out>\n";

    bool binary_write = true;
    MleFaGmmOptions mle_opts;
    std::string update_flags_str = "Uvwb"; 
    ParseOptions po(usage);
    po.Register("binary", &binary_write, "Write output in binary mode");
    po.Register("update-flags", &update_flags_str, "Which Fa-GMM parameters will be "
                "updated: subset of UvwbG.");
    mle_opts.Register(&po);

    po.Read(argc, argv);

    if (po.NumArgs() != 3) {
      po.PrintUsage();
      exit(1);
    }

    std::string model_in_filename = po.GetArg(1),
        stats_filename = po.GetArg(2),
        model_out_filename = po.GetArg(3);

    FaGmm fagmm;
    {
      bool binary_read;
      Input ki(model_in_filename, &binary_read);
      fagmm.Read(ki.Stream(), binary_read);
     // fagmm.InitializeG();
    }

    AccFaGmm fagmm_accs;
    {
      bool binary;
      Input ki(stats_filename, &binary);
      fagmm_accs.Read(ki.Stream(), binary, true /* add accs, doesn't matter */);
    }

    MleFaGmmUpdater updater(mle_opts, StringToFaGmmFlags(update_flags_str));
    {  // Update the FaGmm.
      BaseFloat objf_impr, count;
      count = fagmm_accs.occupancy().Sum();
      objf_impr = updater.Update(fagmm_accs, &fagmm);
      KALDI_LOG << "Overall objective function improvement is "
                << objf_impr << " per frame over "
                <<  (count) <<  " frames.";
    }
    {
      Output ko(model_out_filename, binary_write);
      fagmm.Write(ko.Stream(), binary_write);
      KALDI_LOG << "Written model to " << model_out_filename;
    }
  } catch(const std::exception &e) {
    std::cerr << e.what() << '\n';
    return -1;
  }
}
