

#include "base/kaldi-common.h"
#include "util/common-utils.h"
#include "util/kaldi-io.h"
#include "gmm/diag-gmm.h"
#include "gmm/full-gmm.h"
#include "plda/fa-gmm.h"

int main(int argc, char *argv[]) { 
  try {
        using namespace kaldi;
	
        const char *usage =
        "Initialize the Fa-GMM using a diagonal GMM\n"
        "Usage: fa-gmm-init [options] <gmm-in> <fagmm-out>\n";
    
	bool binary_write = true;
	bool use_full_gmm = false;
	int32 speech_dim = 40;
	int32 state_dim = 40;
	ParseOptions po(usage);
	po.Register("binary", &binary_write, "Write output in binary mode");
	po.Register("speech-dim", &speech_dim, "dimensionality of the speech vectors");
	po.Register("state-dim", &state_dim, "dimensionality of the state vectors");
	po.Register("use-full-gmm", &use_full_gmm, "Initialize from a Full or Diag GMM");
	
	po.Read(argc, argv);

        if (po.NumArgs() != 2) {
          po.PrintUsage();
          exit(1);
        }
    
        std::string gmm_in_filename = po.GetArg(1),
        fagmm_out_filename = po.GetArg(2);
	
	if(use_full_gmm){
	  FullGmm full_gmm;
	  {
	    bool binary_read;
	    Input ki(gmm_in_filename, &binary_read);
	    full_gmm.Read(ki.Stream(), binary_read);
	  }
	
	  FaGmm fa_gmm;
	  fa_gmm.InitFromFullGmm(full_gmm, speech_dim);
	  {
	    Output ko(fagmm_out_filename, binary_write);
	    fa_gmm.Write(ko.Stream(), binary_write);
	  }
	} else {
	  
  	  DiagGmm diag_gmm;
	  {
	    bool binary_read;
	    Input ki(gmm_in_filename, &binary_read);
	    diag_gmm.Read(ki.Stream(), binary_read);
	  }
	
	  FaGmm fa_gmm;
	  fa_gmm.InitFromDiagGmm(diag_gmm, speech_dim, state_dim);
	  {
	    Output ko(fagmm_out_filename, binary_write);
	    fa_gmm.Write(ko.Stream(), binary_write);
	  }
	}
	
  } catch(const std::exception &e) {
    std::cerr << e.what();
    return -1;
  }

}
