//  Liang Lu (2014)
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.

#include "util/common-utils.h"
#include "gmm/full-gmm.h"
#include "gmm/mle-full-gmm.h"
#include "plda/fa-gmm.h"

int main(int argc, char *argv[]) {
  try {
    using namespace kaldi;
    typedef kaldi::int32 int32;

    const char *usage =
        "Convert single factor-analysis GMM to single full-covariance GMM.\n"
        "Usage: fa-gmm-to-fgmm [options] 1.fagmm 1.fgmm\n";
        
    bool binary = true;
    ParseOptions po(usage);
    po.Register("binary", &binary, "Write output in binary mode");
    po.Read(argc, argv);

    if (po.NumArgs() != 2) {
      po.PrintUsage();
      exit(1);
    }

    std::string fagmm_rxfilename = po.GetArg(1),
        fgmm_wxfilename = po.GetArg(2);
    
    FaGmm fagmm;
    
    {
      bool binary_read;
      Input ki(fagmm_rxfilename, &binary_read);
      fagmm.Read(ki.Stream(), binary_read);
    }

    int32 num_gauss = fagmm.NumGauss();
    int32 feature_dim = fagmm.FeatureDim();
    
    FullGmm fgmm(num_gauss, feature_dim);
    Matrix<BaseFloat> means(num_gauss, feature_dim);
    std::vector< SpMatrix<BaseFloat> > inv_covars(num_gauss);
    for(int32 m=0; m<num_gauss; m++) {
      inv_covars[m].Resize(feature_dim);
      means.CopyRowFromVec((Vector<BaseFloat>)fagmm.BiasVector(m), m);
      inv_covars[m].CopyFromSp(fagmm.Inv_Sigma(m));
    }
    
    fgmm.SetWeights(fagmm.weights());
    fgmm.SetInvCovarsAndMeans(inv_covars, means);
    fgmm.ComputeGconsts();
    
    WriteKaldiObject(fgmm, fgmm_wxfilename, binary);
    KALDI_LOG << "Written full GMM to " << fgmm_wxfilename;
  } catch(const std::exception &e) {
    std::cerr << e.what() << '\n';
    return -1;
  }
}

