// pldabin/plda-init.cc

// Liang Lu (2014)

// See ../../COPYING for clarification regarding multiple authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.

#include "util/common-utils.h"
#include "lfa/fa-gmm.h"
#include "plda/am-plda.h"
#include "hmm/transition-model.h"
#include "tree/context-dep.h"


int main(int argc, char *argv[]) {
  try {
    using namespace kaldi;
    typedef kaldi::int32 int32;

    const char *usage =
        "Initialize an AmPlda from a trained factor analysis GMM and a specified"
        " model topology.\n"
        "Usage: plda-init [options] <topology> <tree> <fa-gmm> <plda-out>\n";

    bool binary = true;
//    int32 phn_space_dim = 0, spk_space_dim = 0;
    kaldi::ParseOptions po(usage);
    po.Register("binary", &binary, "Write output in binary mode");

    po.Read(argc, argv);

    if (po.NumArgs() != 4) {
      po.PrintUsage();
      exit(1);
    }

    std::string topo_in_filename = po.GetArg(1),
        tree_in_filename = po.GetArg(2),
        init_model_filename = po.GetArg(3),
        plda_out_filename = po.GetArg(4);

    ContextDependency ctx_dep;
    {
      bool binary_in;
      Input ki(tree_in_filename.c_str(), &binary_in);
      ctx_dep.Read(ki.Stream(), binary_in);
    }


    HmmTopology topo;
    ReadKaldiObject(topo_in_filename, &topo);

    TransitionModel trans_model(ctx_dep, topo);

    kaldi::AmPlda plda;
    kaldi::FaGmm fagmm;
    {
      bool binary_read;
      kaldi::Input ki(init_model_filename, &binary_read);
      fagmm.Read(ki.Stream(), binary_read);
    }
    plda.InitializeFromFaGmm(fagmm, trans_model.NumPdfs());
    
 //   plda.ComputeNormalizers();

    {
      kaldi::Output ko(plda_out_filename, binary);
      trans_model.Write(ko.Stream(), binary);
      plda.Write(ko.Stream(), binary, kaldi::kPldaWriteAll);
    }

    KALDI_LOG << "Written model to " << plda_out_filename;
  } catch(const std::exception &e) {
    std::cerr << e.what() << '\n';
    return -1;
  }
}


